# -*- coding: utf-8 -*-
"""
Created on Mon Mar  7 11:08:29 2022

@author: Stian Valand
"""

import numpy as np
import matplotlib.pyplot as plt 
np.set_printoptions(suppress=True)
import warnings as wa
import time
import pickle


wa.filterwarnings('ignore')


#%% Some parameters
x_start = 0; x_end = 45
L = x_end - x_start #Length of domain
t_start = 0; t_end = 0.01
Nx = 2000 # Number of spatial iterations
dx = ((x_end - x_start)/(Nx))  # Spatial step
x = np.linspace(x_start + dx/2, x_end, Nx+1) # For plotting

start_time = time.time()

half_way = int(Nx/2)
Nt = 1350  #Number of time-steps
leak_time = 300#12450 # Onset leakage 

#%% Pipe specs.
D = 0.009 #pipe diameter
OD = 0.012 # Outer diameter
ed = (0.025*1e-3)/D #Relative roughness. Steel structural or forged
K = 0.3 # Minor loss coefficient 90deg regular elbow
N_bend = 20 #Number of bends
Kt = K/dx #(N_bend * K) # Total number of bends
A = (np.pi/4.0) * D**2 # Cross-sectional area of the pupe

Bloc = int((Nx-(2*(Nx/N_bend)))/N_bend) #Location of pipe bends. 
#%% Leak specs
leak = int(Nx/2) #leak location
Do = 0.001 # Leak orifice diameter 
Ao = (np.pi/4.0) * Do**2 # leak orifice area

#%% Parameters
pa = 101325 # Pa, atmospheric pressure
R = 4124.2 # universal gas constant, Hydrogen
gamma = 1.41 # Cp/Cv
b = 7.691e-3 # Abel-Noble co-volume constant [m3/kg]
CPR = ((gamma+1)/2)**(gamma / (gamma - 1.0)) # Critical pressure ratio, inside to outside ratio
k_pipe = 45 # Thermal conductivity pipeline material [W/mK]
h_air = 50 # Convective heat transfer coefficient[W/m^2*K], assumed number.
Cp = 0.0144 # Specific heat capacity [J/kg*K]
#%% Arrays
u = np.zeros(Nx+1)
p = np.zeros(Nx+1)
T = np.zeros(Nx+1)
rho = np.zeros(Nx+1)
E = np.zeros(Nx+1)
C = np.zeros(Nx+1)
Ta = np.ones(Nx+1)*293.15

lostMass = np.zeros(Nx+1)
lostMass1 = np.zeros(Nx+1)
Mdl0 = np.zeros(Nx+1)
Mdl1 = np.zeros(Nx+1)
Mdl2 = np.zeros(Nx+1)

#%% Courant-Friedrich-Lewy number
CFL = 0.87

#%% Flux limitor selector. SUPERBEE = 1, MINBEE = 2
phip = phim = 2

#%% IC selector 
ic = 4        #(IC 4 & 5 needs to be used together with "steady state" files which must be stored in same folder as simulation file. If IC 3 is to be used with ideal gas, 'b' must manually be set to zero in the parameter section)
#%% Initial Conditions 
if ic == 1: # Test 1, Sod's problem, Set Nt = 300, leak_time = 0, Blank out line 399 - 410.
    
    u[:half_way] = 0.0 
    rho[:half_way] = 1.0 
    p[:half_way] = 1.0
    u[half_way:] = 0.0 
    rho[half_way:] = 0.125
    p[half_way:] = 0.1
    
elif ic == 2: #Test 2, Set Nt = 200, leak_time = 0, Blank out line 399 - 410.

    u[:half_way] = -2.0 
    rho[:half_way] = 1.0 
    p[:half_way] = 0.4
    u[half_way:] = 2.0 
    rho[half_way:] = 1.0
    p[half_way:] = 0.4

elif ic == 3: # Initial conditions. Should be run until steady state. E.g, Nt = 13500, leak_time = 12450.

    u[:] = 30 #[m/s]
    rho[:] = 0.822 #[kg/m3]
    p[:] = 10e5 #[Pa]

elif ic == 4: # Abel-Noble EOS. Nt = 1350, leak_time = 300, Must be used with the file "steady". Simulations in steady-state conditions with Abel-Noble EOS.
    Steady = pickle.load(open('steady', 'rb')) 
    u[:] = Steady[0]
    rho[:] = Steady[1] 
    p[:] = Steady[2]

elif ic == 5: # Ideal gas EOS. Nt = 1350, leak_time = 300, Must be used with the file "steady_Ideal_Gas". Simulations in steady-state conditions with Ideal gas.
    Steady_IG = pickle.load(open('steady_Ideal_Gas', 'rb')) 
    u[:] = Steady_IG[0]
    rho[:] = Steady_IG[1] 
    p[:] = Steady_IG[2]


#%% Some initial conditions
# Total energy
if ic == 4: 
    
    E = Steady[3]
    
elif ic == 5:
    E = Steady_IG[3]
    b = 0
    
else:
    E = p*(1-b*rho)/(gamma-1) + 0.5 * u**2 * rho

# Temperature

T = (p*(1-(b*rho)))/(R * rho) 

#Speed of sound
C = ((gamma * p) / ((1.0 - b*rho)*rho))**0.5 

# Mass flow rate [kg/s]
m_dot = rho * u * A 

#%% Some isentropic relations
# Sonic Relations

rhoe = rho * (2.0/(gamma +1)) ** (1.0/(gamma - 1.0))
pe = p * (2.0 / (gamma + 1.0))**(gamma / (gamma - 1.0))
ue = C * ((2.0 / (gamma + 1))**0.5)
Ee = pe/(gamma-1) + 0.5 * (ue**2) * rhoe

massLeak1 =  ( rhoe * ue * (Do/D)**2)/dx

impLeak1 = ( u * massLeak1)

enerLeak1 = ( Ee + pe) * massLeak1 /rhoe
massOut1 = rhoe * ue * Ao

# Subsonic relations
rhoSub = ((pa/p)**(1.0/gamma)) * rho
uSub = C * ( (2.0 / (gamma - 1.0) ) * (((pa / p)**((gamma - 1) / gamma))-1))**0.5
ESub = (pa/(gamma - 1)) + 0.5 * rhoSub * uSub**2

  
massLeak2 =  (( rhoSub * uSub )*(Do/D)**2)/dx

impLeak2 = (u * massLeak2)

enerLeak2 = (ESub + pa) * massLeak2/rhoSub
massOut2 = rhoSub * uSub * Ao

#%% Timestep
dt = (CFL*dx)/(max(abs(u)+C))


#%% Linear interpolation for the viscosity calculation
def lin_interpol(y1, y2, x_r, x1, x2): 
    y = np.zeros(Nx+1)
    y = ((y2-y1)/((x2-x1))*(x_r-x1))+y1
  
    return y

#%% Friction factor, viscosity and Reynoldsnumber calculation:
def friction(eD, rho, u, D, T, p):

    #Churchills correlation for friction factor. Valid for all numbers of Re.
    #Dynamic viscosity (Mu) from empirical data, @ 1, 10, 50 and 100 bar. Linear interpolation between the pressures.    
    
    Mu1 = ((0.0217 * T) + 2.2396)*1e-6 # 1 bar
    Mu10 = ((0.0216 * T) + 2.4289)*1e-6 # 10 bar
    Mu50 = ((0.0204 * T) + 2.8958)*1e-6 # 50 bar
    Mu100 = ((0.0196 * T) + 3.2723)*1e-6 # 100 bar
    
    
    p1 = np.select([(p >= 1e5) & (p < 10e5) , (p >= 10e5) & (p < 50e5), (p >= 50e5) & (p < 100e5)], [1e5, 10e5, 50e5])
    p2 = np.select([(p >= 1e5) & (p < 10e5) , (p >= 10e5) & (p < 50e5), (p >= 50e5) & (p < 100e5)], [10e5, 50e5, 100e5])
    
    Mu11 = np.select([(p >= 1e5) & (p < 10e5) , (p >= 10e5) & (p < 50e5), (p >= 50e5) & (p < 100e5)], [Mu1, Mu10, Mu50])
    Mu22 = np.select([(p >= 1e5) & (p < 10e5) , (p >= 10e5) & (p < 50e5), (p >= 50e5) & (p < 100e5)], [Mu10, Mu50, Mu100])
    
    Mu = np.select([p < 1e5, p >= 100e5], [Mu1, Mu100], default = lin_interpol(Mu11, Mu22, p, p1, p2))
  
    #Reynolds number
    Re = (rho*abs(u)*D)/(Mu)
    
    #Churchills correlation:

    Ac = (2.457 * np.log(1.0/((ed*0.27)  + (7.0/Re)**0.9)))**16.0
        
    Bc = (37530.0 / Re)**16.0
    
    f = 8.0*(((8.0/Re)**12.0) +(Ac + Bc)**(-1.5))**(1.0/12.0)      
    
    f[Re <= 100.0] = 0.6
     
    return(f, Re, Mu) #Darcy friction factor, Reynolds number, Dynamic Viscosity

#%% Thermal conductivity of Hydrogen
def thermalcondH2(T, p):
    k1 = ((0.5387 * T) + 22.433)*1e-3 # 1 bar
    k10 = ((0.5352 * T) + 24.298)*1e-3 # 10 bar
    k50 = ((0.501 * T) + 38.332)*1e-3 # 50 bar
    k100 = ((0.4784 * T) + 49.584)*1e-3 # 100 bar
       
    p1 = np.select([(p >= 1e5) & (p < 10e5) , (p >= 10e5) & (p < 50e5), (p >= 50e5) & (p < 100e5)], [1e5, 10e5, 50e5])
    p2 = np.select([(p >= 1e5) & (p < 10e5) , (p >= 10e5) & (p < 50e5), (p >= 50e5) & (p < 100e5)], [10e5, 50e5, 100e5])
    
    k11 = np.select([(p >= 1e5) & (p < 10e5) , (p >= 10e5) & (p < 50e5), (p >= 50e5) & (p < 100e5)], [k1, k10, k50])
    k22 = np.select([(p >= 1e5) & (p < 10e5) , (p >= 10e5) & (p < 50e5), (p >= 50e5) & (p < 100e5)], [k10, k50, k100])
    
    k = np.select([p < 1e5, p >= 100e5], [k1, k100], default = lin_interpol(k11, k22, p, p1, p2))
    
    return(k)

#%% Heat transfer
def heatSource(Re, T, p, Ta, L, k_pipe, h_air, Cp, Mu, D, OD):
    r1 = D / 2
    r2 = OD / 2
    
    kH2 = thermalcondH2(T , p)
    
    n = np.select([T <= Ta, T > Ta], [0.3, 0.4])
    Pr = Mu * Cp / kH2
    Nu = 0.023 * (Re**0.8) * Pr**n #Dittus - Boelter correlation
    
    h_gw = Nu * kH2/L
    
    U = ((1.0 / h_gw) + (r1 * (np.log(r2 / r1)) / k_pipe) + ((r1 /(r2 * h_air)) ))**-1
    
    q = (U * (T - Ta))
    
    return(q)

#%% LAX Wendroff finite difference scheme

def LaxFlux(dt, dx, rho, u,E, p, CFL, f, gamma, C1_fric, C1_minor, C2_heat, C2_fricWork, b, T, phip, phim, Mdl0, Mdl1, Mdl2, \
            massLeak1, enerLeak1, impLeak1, massLeak2, enerLeak2, impLeak2, Do, A, pe, pa, rhoe, Ee, R, ue, m_dot, Ao, massOut1, massOut2, rhoSub, uSub, ESub):
    
    A0 = (rho) 
    A1 = (rho*u)
    A2 = (E) 
    
    #Fluxes, plus(p) and minus(m)
    
    B0 = (rho*u)
    B1 = (p + (rho*u**2))
    B2 = ((E + p)*u)
      
 #%% Richtmeyer   
    # Half step Richtmeyer  
    A0_RI_plus = 0.5*(A0[2:] + A0[1:-1]) - 0.5*(dt/dx)*(B0[2:] - B0[1:-1])
    A1_RI_plus = 0.5*(A1[2:] + A1[1:-1]) - 0.5*(dt/dx)*(B1[2:] - B1[1:-1]) 
    A2_RI_plus = 0.5*(A2[2:] + A2[1:-1]) - 0.5*(dt/dx)*(B2[2:] - B2[1:-1]) 
    
    A0_RI_minus = 0.5*(A0[1:-1] + A0[:-2]) - 0.5*(dt/dx)*(B0[1:-1] - B0[:-2])
    A1_RI_minus = 0.5*(A1[1:-1] + A1[:-2]) - 0.5*(dt/dx)*(B1[1:-1] - B1[:-2]) 
    A2_RI_minus = 0.5*(A2[1:-1] + A2[:-2]) - 0.5*(dt/dx)*(B2[1:-1] - B2[:-2])        
        
    # Compute variables for next step    
        
    # PLus variables Richtmeyer
        
    rho_pR = A0_RI_plus
    u_pR = A1_RI_plus/rho_pR
    E_pR = A2_RI_plus
    p_pR = ((gamma - 1) * (E_pR - (0.5 * rho_pR * u_pR**2)))/(1-b*rho_pR)
        
    #Minus variables Richtmeyer
    rho_mR = A0_RI_minus
    u_mR = A1_RI_minus/rho_mR
    E_mR = A2_RI_minus
    p_mR = ((gamma - 1) * (E_mR - (0.5 * rho_mR * u_mR**2)))/(1-b*rho_mR)
    
    # Compute 1/2 step Richtmeyer fluxes 
    B0_RI_plus = (rho_pR * u_pR)
    B1_RI_plus = (p_pR + (rho_pR * u_pR**2))
    B2_RI_plus = ((E_pR + p_pR) * u_pR)   
    
    B0_RI_minus = (rho_mR* u_mR)
    B1_RI_minus = (p_mR + (rho_mR * u_mR**2))
    B2_RI_minus = ((E_mR + p_mR) * u_mR)   

    #%% Lax-Friedrich (LF) Fluxes
       
    # Compute LF fluxes
    B0_LF_plus = 0.5*(B0[2:] + B0[1:-1]) - 0.5*(dx/dt)*(A0[2:] - A0[1:-1])
    B1_LF_plus = 0.5*(B1[2:] + B1[1:-1]) - 0.5*(dx/dt)*(A1[2:] - A1[1:-1])
    B2_LF_plus = 0.5*(B2[2:] + B2[1:-1]) - 0.5*(dx/dt)*(A2[2:] - A2[1:-1])
       
    B0_LF_minus = 0.5*(B0[1:-1] + B0[:-2]) - 0.5*(dx/dt)*(A0[1:-1] - A0[:-2])
    B1_LF_minus = 0.5*(B1[1:-1] + B1[:-2]) - 0.5*(dx/dt)*(A1[1:-1] - A1[:-2])
    B2_LF_minus = 0.5*(B2[1:-1] + B2[:-2]) - 0.5*(dx/dt)*(A2[1:-1] - A2[:-2])

    #%% Force fluxes

    #Force Plus    
    B0_FORCE_plus = 0.5 * (B0_RI_plus + B0_LF_plus)
    B1_FORCE_plus = 0.5 * (B1_RI_plus + B1_LF_plus)
    B2_FORCE_plus = 0.5 * (B2_RI_plus + B2_LF_plus)

    #Force minus
    B0_FORCE_minus = 0.5 * (B0_RI_minus + B0_LF_minus)
    B1_FORCE_minus = 0.5 * (B1_RI_minus + B1_LF_minus)    
    B2_FORCE_minus = 0.5 * (B2_RI_minus + B2_LF_minus)

    #%% Flux limiter. B_LO = B_Force. B_HI = B_RI. Using q = total energy (E)

    q = E[1:-1]

    q_plus_1 = np.roll(q, -1)
    q_plus_2 = np.roll(q_plus_1, -1)
    q_minus_1 = np.roll(q, 1)
    q_minus_2 = np.roll(q_minus_1, 1)
       
    
    np.seterr(divide='ignore', invalid = 'ignore')
 
    #%% rR Plus
    rRp = (q_plus_2 - q_plus_1) / (q_plus_1 - q)
    rRp = np.nan_to_num(rRp)
    #%% rL plus
    rLp = (q - q_minus_1) / (q_plus_1 - q)
    rLp = np.nan_to_num(rLp)
    #print(rRp)
    #%% Phi LR plus
    phiLRp = np.minimum(rRp, rLp)
    
    
    phiGp = (1.0-CFL)/(1.0+CFL)
    
    #%% rR minus 
    rRm = (q_plus_1 - q) / (q - q_minus_1)
    rRm = np.nan_to_num(rRm)
    
    #%% rL minus 
    rLm = (q_minus_1 - q_minus_2) / (q - q_minus_1)
    rLm = np.nan_to_num(rLm)
    
    #%% Phi LR minus
    phiLRm = np.minimum(rRm, rLm)
    
    
    phiGm = (1.0-CFL)/(1.0+CFL)


    #%% Super-Bee
    
    GpS = (phiGp + (1.0-phiGp) * phiLRp )
    phiSBp = np.select([phiLRp <= 0, phiLRp <= 0.5, phiLRp<=1.0, phiLRp>1.0], [0, 2.0*phiLRp, 1.0, np.minimum(2.0, GpS) ])

  
    GmS = (phiGm + (1.0-phiGm) * phiLRm )
    phiSBm = np.select([phiLRm <= 0, phiLRm <=0.5, phiLRm <=1.0, phiLRm > 1.0], [0, 2.0*phiLRm, 1.0, np.minimum(2.0, GmS)])
    
    #%% MINBEE
    phiMBp = np.select([phiLRp <= 0, phiLRp <= 1.0, phiLRp>1], [0, phiLRp, 1.0])
    
    phiMBm = np.select([phiLRm <= 0, phiLRm <= 1.0, phiLRm>1], [0, phiLRm, 1.0])
    
    #%% Flux limitor selection
    phiP = np.select([ phip == 1, phip == 2], [phiSBp, phiMBp])
    phiM = np.select([ phim == 1, phim == 2], [phiSBm, phiMBm])
      
    #%% Flux i+1/2
    B0_half_plus = B0_FORCE_plus +  phiP*(B0_RI_plus - B0_FORCE_plus)
    B1_half_plus = B1_FORCE_plus +  phiP*(B1_RI_plus - B1_FORCE_plus)
    B2_half_plus = B2_FORCE_plus + phiP*(B2_RI_plus - B2_FORCE_plus)
    
    #%% Flux i-1/2
    B0_half_minus = B0_FORCE_minus + phiM*(B0_RI_minus - B0_FORCE_minus)
    B1_half_minus = B1_FORCE_minus + phiM*(B1_RI_minus - B1_FORCE_minus)
    B2_half_minus = B2_FORCE_minus + phiM*(B2_RI_minus - B2_FORCE_minus)

    #%% Move one time step

    A0_new = A0[1:-1] - (dt/dx)*(B0_half_plus - B0_half_minus)
    A1_new = A1[1:-1] - (dt/dx)*(B1_half_plus - B1_half_minus)
    A2_new = A2[1:-1] - (dt/dx)*(B2_half_plus - B2_half_minus)
    
    #%% Source terms
    
    A1_new[1:-1] = A1_new[1:-1] + (dt*C1_fric[1:-1])
    A2_new[1:-1] = A2_new[1:-1] + (dt * C2_heat[1:-1]) + (dt * C2_fricWork[1:-1])
    
    A1_new[[Bloc, 2*Bloc, 3*Bloc, 4*Bloc, 5*Bloc, 6*Bloc, 7*Bloc, 8*Bloc, 9*Bloc, 10*Bloc, 11*Bloc, 12*Bloc, 13*Bloc, 14*Bloc, 15*Bloc, 16*Bloc, 17*Bloc, 18*Bloc, 19*Bloc, 20*Bloc]] = \
       A1_new[[Bloc, 2*Bloc, 3*Bloc, 4*Bloc, 5*Bloc, 6*Bloc, 7*Bloc, 8*Bloc, 9*Bloc, 10*Bloc, 11*Bloc, 12*Bloc, 13*Bloc, 14*Bloc, 15*Bloc, 16*Bloc, 17*Bloc, 18*Bloc, 19*Bloc, 20*Bloc]] +\
             (dt * C1_minor)
   
  
   #%% Leakage      
    A0_new[leak] = A0_new[leak] + (dt * Mdl0[leak])
    A1_new[leak] = A1_new[leak] + (dt * Mdl1[leak])
    A2_new[leak] = A2_new[leak] + (dt * Mdl2[leak])
    #%% Calculate new variables

    rho_new = A0_new
    u_new = A1_new/rho_new
    E_new = A2_new
    p_new = ((gamma - 1) * (E_new - (0.5 * rho_new * u_new**2)))/(1-b*rho_new)
    T_new = (p_new*(1-(b*rho_new)))/(R * rho_new) 
    
    C_new = ((gamma * p_new) / ((1.0 - b*rho_new)*rho_new))**0.5
    
    m_dot_new = rho_new * u_new * A

    #%% Sonic relations
    
    pe_new = p_new * (2.0 / (gamma + 1.0))**(gamma / (gamma - 1.0))
    rhoe_new = rho_new * (2.0/(gamma +1)) ** (1.0/(gamma - 1))

    ue_new = C_new * ((2.0 / (gamma +1))**0.5)
    
    Ee_new = (pe_new / (gamma -1.0)) + 0.5 * (ue_new**2) * rhoe_new
    
    massLeak1_new =  (( rhoe_new * ue_new )*(Do/D)**2)/dx

    impLeak1_new = (u_new * massLeak1_new)

    enerLeak1_new = ( Ee_new + pe_new) * massLeak1_new/rhoe_new
    
    massOut1_new = rhoe_new * ue_new * Ao
    
    #%% Subsonic relations
    rhoSub_new = (pa/p_new)**(1.0/gamma) * rho_new
    
    uSub_new = C_new * (( (2.0 / (gamma - 1.0) ) * (((pa / p_new)**((gamma - 1) / gamma))-1))**0.5)
    
    ESub_new = (pa/(gamma - 1)) + 0.5 * rhoSub_new * uSub_new**2
    
    massLeak2_new =  ( rhoSub_new * uSub_new )*((Do/D)**2)/dx

    impLeak2_new = (u_new * massLeak2_new)

    enerLeak2_new = ( ESub_new + pa) * massLeak2_new/rhoSub_new
    
    massOut2_new = rhoSub_new * uSub_new * Ao
    
    
    
    return(rho_new, u_new, E_new, p_new, C_new, T_new, m_dot_new, massLeak1_new, enerLeak1_new, impLeak1_new,\
          massLeak2_new, enerLeak2_new, impLeak2_new, pe_new, rhoe_new, ue_new, Ee_new, uSub_new, rhoSub_new, ESub_new, massOut1_new, massOut2_new)


#%% Arrays for pressure sensors and mass flow plotting
ps1 = []
ps2 = []
ps3 = []
ps4 = []

mass1 = []
mass2 = []
mass3 = []
mass4 = []
massLekk = []
tid = []
t = 0


#%% Time loop
it = 0
for n in range(0, Nt): 
    
    
    # Friction term
    (f, Re, Mu) = friction(ed, rho[1:-1], u[1:-1], D, T[1:-1], p[1:-1])
    
    # Determine sonic or subsonic flow at leakage point
    if  p[leak]/pa >= CPR:
        massLeak = massLeak1
        impLeak = impLeak1
        enerLeak = enerLeak1
        massOut = massOut1    
    else:
        massLeak = massLeak2
        impLeak = impLeak2
        enerLeak = enerLeak2
        massOut = massOut2
        
    # Initiate leakage
    if n < leak_time:
        Mdl0[leak] = 0
        Mdl1[leak] = 0
        Mdl2[leak] = 0
        lostMass[leak] = 0
    else:
        Mdl0[leak] = -massLeak[leak]
        Mdl1[leak] = -impLeak[leak]
        Mdl2[leak] = -enerLeak[leak]
        lostMass[leak] = - massOut[leak]
    
    
    #print(massOut)
    #Source terms:
    C1_fric = - ((f*rho[1:-1]*abs(u[1:-1])*u[1:-1])/(2*D)) # Source term due to internal friction in the pipe
   
    C1_minor = - (Kt * rho[[Bloc, 2*Bloc, 3*Bloc, 4*Bloc, 5*Bloc, 6*Bloc, 7*Bloc, 8*Bloc, 9*Bloc, 10*Bloc, 11*Bloc, 12*Bloc, 13*Bloc, 14*Bloc, 15*Bloc, 16*Bloc, 17*Bloc, 18*Bloc, 19*Bloc, 20*Bloc]] * \
                  u[[Bloc, 2*Bloc, 3*Bloc, 4*Bloc, 5*Bloc, 6*Bloc, 7*Bloc, 8*Bloc, 9*Bloc, 10*Bloc, 11*Bloc, 12*Bloc, 13*Bloc, 14*Bloc, 15*Bloc, 16*Bloc, 17*Bloc, 18*Bloc, 19*Bloc, 20*Bloc]]**2)/2.0 # Source term due to bend
    
    C2_heat =  - 4*heatSource(Re, T[1:-1], p[1:-1], Ta[1:-1], L, k_pipe, h_air, Cp, Mu, D, OD)/D # Heat transfer term
   
    C2_fricWork = - (2.0 * f * rho[1:-1] * abs(u[1:-1])**3) / D # Friction work term
    
      
    # Initiate Richtmeyer/Friedrich FLIC scheme for computation of variables
    (rho_new, u_new, E_new, p_new, C_new, T_new, m_dot_new, massLeak1_new, enerLeak1_new, impLeak1_new,\
     massLeak2_new, enerLeak2_new, impLeak2_new, pe_new, rhoe_new, ue_new, Ee_new, uSub_new, rhoSub_new, ESub_new, massOut1_new, massOut2_new ) = LaxFlux(dt, dx, rho, u, E, p, CFL, f, gamma, C1_fric, C1_minor, C2_heat, C2_fricWork, b, T, phip, phim, Mdl0, Mdl1, Mdl2, \
                                                                              massLeak1, enerLeak1, impLeak1, massLeak2, enerLeak2, impLeak2, Do, A, pe, pa, rhoe, Ee, R, ue, m_dot, Ao, massOut1, massOut2, rhoSub, uSub, ESub)
                                                                                  
    p_new[0] = p[0] # Fixed Value                                                                                  
    u_new[0] = u_new[1] # ZeroGradient
    rho_new[0] = rho_new[1] # ZeroGradient
    E_new[0] = E_new[1] # ZeroGradient
    T_new[0] = T[0] #Zero Gradient
   
    
    p_new[-1] = p_new[-2] #Zero Gradient
    rho_new[-1] = rho_new[-2] #ZeroGradient
    E_new[-1] = E_new[-2] #Zero Gradient
    T_new[-1] = T_new[-2] #Zero Gradient                                                                                   
    u_new[-1] = u[-1]  # Fixed value

    
    #Update variables
    u[1:-1] = u_new
    rho[1:-1] = rho_new
    E[1:-1] = E_new
    p[1:-1] = p_new
    T[1:-1] = T_new
    C[1:-1] = C_new
    pe[1:-1] = pe_new
    rhoe[1:-1] = rhoe_new
    uSub[1:-1] = uSub_new
    Ee[1:-1] = Ee_new
    rhoSub[1:-1] = rhoSub_new
    ue[1:-1] = ue_new
    ESub[1:-1] = ESub_new
    m_dot[1:-1] = m_dot_new
    
    massLeak1[1:-1] = massLeak1_new
    impLeak1[1:-1] = impLeak1_new
    enerLeak1[1:-1] = enerLeak1_new
    massOut1[1:-1] = massOut1_new  
    massLeak2[1:-1] = massLeak2_new
    impLeak2[1:-1] = impLeak2_new
    enerLeak2[1:-1] = enerLeak2_new
    massOut2[1:-1] = massOut2_new 
    
    dt = (CFL*dx)/(max((abs(u[1:-1])+C[1:-1]))) #Update time step
    
    ps1.append(p[333])
    ps2.append(p[667])
    ps3.append(p[1334])
    ps4.append(p[1667])
    
    mass1.append(m_dot[333])
    mass2.append(m_dot[667])
    mass3.append(m_dot[1334])
    mass4.append(m_dot[1667])
   
    massLekk.append(lostMass[leak])       
    
   
    
    if it%50 == 0:
        print (it)
       # fig,axes = plt.subplots(nrows=4, ncols=1)
       # plt.subplot(4, 1, 1)
       # plt.plot(x[1:-1], rho[1:-1], 'k-')
       # plt.ylabel('$rho$',fontsize=16)
       # plt.xlim(x_start,x_end)
       # plt.tick_params(axis='x',bottom=False,labelbottom=False)
       # plt.grid(True)

       # plt.subplot(4, 1, 2)
       # plt.plot(x[1:-1], u[1:-1], 'r-')
       # plt.ylabel('$U$',fontsize=16)
       # plt.xlim(x_start,x_end)
       # plt.tick_params(axis='x',bottom=False,labelbottom=False)
       # plt.grid(True)

       # plt.subplot(4, 1, 3)
       # plt.plot(x[1:-1], p[1:-1], 'b-')
       # plt.ylabel('$p$',fontsize=16)
       # plt.xlim(x_start,x_end)
       # plt.tick_params(axis='x',bottom=False,labelbottom=False)
       # plt.grid(True)
    
        #plt.subplot(4, 1, 4)
        #plt.plot(x[1:-1], E[1:-1], 'g-')
        #plt.ylabel('$E$',fontsize=16)
        #plt.grid(True)
        #plt.xlim(x_start,x_end)
        #plt.xlabel('x',fontsize=16)
        #plt.subplots_adjust(left=0.2)
        #plt.subplots_adjust(bottom=0.15)
        #plt.subplots_adjust(top=0.95)
        #plt.show()
    it = it + 1
    t = t + dt
    tid.append((t*1e3))

    
print('Start Leak:', tid[leak_time],'ms, and after:', leak_time, 'timesteps') 
print('The simulation is finished after:', tid[n], 'ms which corresponds to', n+1, 'time steps')   
end_time = time.time()

plt.subplot(4, 1, 1)
plt.plot(x[1:-1], rho[1:-1], 'b-')
plt.ylabel('Density [kg/m3]')
plt.title('Flow variables 10 bar, 1.00 mm leak diameter @ 22.5 m')
plt.grid(True)
plt.tick_params(axis='x',bottom=False,labelbottom=False)
plt.xlim(x_start,x_end)

plt.subplot(4, 1, 2)
plt.plot(x[1:-1], u[1:-1], 'g-')
plt.ylabel('Velocity [m/s]')
plt.grid(True)
plt.xlim(x_start,x_end)
plt.tick_params(axis='x',bottom=False,labelbottom=False)

plt.subplot(4, 1, 3)
plt.plot(x[1:-1], p[1:-1], 'r-')
plt.ylabel('Pressure [Pa]')
plt.grid(True)
plt.xlim(x_start,x_end)
plt.tick_params(axis='x',bottom=False,labelbottom=False)

plt.subplot(4, 1, 4)
plt.plot(x[1:-1], E[1:-1], 'k-')
plt.ylabel('Total Energy [J]')
plt.grid(True)
plt.xlim(x_start,x_end)
plt.show()

#plt.xaxis.set_major_formatter(FormatStrFormatter('%.2f'))
plt.plot(tid, ps1, 'b', label = "PS1 @ 7.5m")
plt.plot(tid, ps2, 'r', label = "PS2 @ 15m")
plt.plot(tid, ps3, 'g', label = "PS3 @ 30m")
plt.plot(tid, ps4, 'k',label = "PS4 @ 37.5m")
plt.ylabel('Pressure [Pa]')
plt.xlabel('Time [ms]')
plt.legend(loc = "lower left")
plt.title(' Pressure drop, Leak: 1.00 mm, leak location: 22.5m, leak start: t = 4.4 ms')
plt.tick_params(axis='x', which='major', labelsize=8)
plt.ticklabel_format( style = 'plain')
plt.show()

plt.plot(tid, mass1, 'b', label = "Sensor 1 @ 7.5m")
plt.plot(tid, mass2, 'r', label = "Sensor 2 @ 15m")
plt.plot(tid, mass3, 'g', label = "Sensor 3 @ 30m")
plt.plot(tid, mass4, 'k',label = "Sensor 4 @ 37.5m")
#plt.plot(tid, massLekk, 'm', label = 'Mass flow leakage')
plt.ylabel('Mass Flow [kg/s]')
plt.xlabel('Time [ms]')
plt.legend(loc = "best")
plt.title('Mass Flow, Leak: 1.00 mm, location = 22.5m, leak start: t = 4.4 ms')
plt.tick_params(axis='x', which='major', labelsize=8)
plt.ticklabel_format( style = 'plain')
plt.show()

plt.plot(tid, massLekk, 'b', label = 'Mass flow leakage')
plt.title('Mass flow leakage, Leak: 1.00 mm, leak location = 22.5m')
plt.ylabel('Mass flow [kg/s]')
plt.legend(loc = "best")
plt.xlabel('Time [ms]')
plt.ticklabel_format(style = 'plain')
plt.show()

plt.plot(tid, mass1, 'r', label = 'Mass flow @ PS1')
plt.title('Mass flow Sensor 1: 1.00 mm, leak location = 22.5m')
plt.ylabel('Mass flow [kg/s]')
plt.legend(loc = "best")
plt.xlabel('Time [ms]')
plt.ticklabel_format(style = 'plain')
plt.show()

plt.plot(tid[1000:1300], ps1[1000:1300], 'b',label = "PS1 @ 7.5m")
plt.ylabel('Pressure [Pa]')
plt.xlabel('Time [ms]')
plt.legend(loc = "lower left")
plt.title(' Pressure drop PS1, Leak: 1.00 mm, leak location = 22.5m, leak start = 4.4 ms')
plt.ticklabel_format(style = 'plain')
plt.show()

plt.plot(tid[660:800], ps2[660:800], 'r',label = "PS2 @ 15m")
plt.ylabel('Pressure [Pa]')
plt.xlabel('Time [ms]')
plt.legend(loc = "lower left")
plt.title(' Pressure drop PS2, Leak: 1.00 mm, leak location = 22.5m, leak start = 4.4 ms')
plt.ticklabel_format(style = 'plain')
plt.show()


plt.plot(tid[660:800], ps3[660:800], 'g',label = "PS3 @ 30m")
plt.ylabel('Pressure [Pa]')
plt.xlabel('Time [ms]')
plt.legend(loc = "lower left")
plt.title(' Pressure drop PS3, Leak: 1.00 mm, leak location = 22.5m, leak start = 4.4 ms')
plt.ticklabel_format(style = 'plain')
plt.show()

plt.plot(tid[1000:1300], ps4[1000:1300], 'k',label = "PS4 @ 37.5m")
plt.ylabel('Pressure [Pa]')
plt.xlabel('Time [ms]')
plt.legend(loc = "lower left")
plt.title(' Pressure PS4, Leak: 1.00 mm, leak location = 22.5m, leak start = 4.4 ms')
plt.ticklabel_format(style = 'plain')
plt.show()



print('Mass flow inlet =', mass1[1])
print('LekkasjeFlow=', (max((massLekk))- min((massLekk))))
print('Change in Mass1 =', (max(mass1[1000:1200]) - min(mass1[1000:1200])))

print('Change in Mass2 =', (max(mass2[660:760]) - min(mass2[660:760])))

print('Change in Mass3 =', (max(mass3[660:760]) - min(mass3[660:760])))

print('Change in Mass4 =', (max(mass4[1000:1200]) - min(mass4[1000:1200])))

print('Delta P, PS1 =', (max(ps1[1000:1150]) - min(ps1[1000:1150])))

print('Delta P, PS2 =', (max(ps2[660:760]) - min(ps2[660:760])))

print('Delta P, PS3 =', (max(ps3[660:760]) - min(ps3[660:760])))

print('Delta P, PS4 =', (max(ps4[1000:1150]) - min(ps4[1000:1150])))

print('Total pressure drop, inlet to outlet', (max(p) - min(p)))


#np.savetxt('leak_1mm.csv', [u, p, rho, E], delimiter=',', fmt='%s')
  

